/*
 * LBP Detector Toolbox
 * Roman Juranek <ijuranek@fit.vutbr.cz>
 * Faculty of Information Technology, BUT, Brno
 *
 */

#include "mex.h"
#include <vector>
#include <list>
#include <cmath>
#include <numeric>
#include <algorithm>
#include <ThreadPool.h>
#include <common.h>
#include "lbpdetector.h"

using namespace std;
using namespace LBPDetector;

// Create C structure from parameters obtained from MATLAB.
Detector * createDetectorFromMxArray(const mxArray * ftr, const mxArray * alpha, const mxArray * dim, const mxArray * cascThr)
{
    const int * sz = (const int*)mxGetData(dim); // 1 x 2 int
    Detector * D = createDetector(sz[0], sz[1], mxGetN(ftr), (unsigned char*)mxGetData(ftr), (float*)mxGetData(alpha), (float)mxGetScalar(cascThr));
    return D;
}

void mexFunction(int nlhs, mxArray ** lhs, int nrhs, const mxArray ** rhs)
{
    if (nrhs < 6)
    {
        mexErrMsgTxt("Error calling mexScalarDetect");
    }

    const mxArray * mx_imgs = rhs[0];
    const mxArray * mx_scales = rhs[1];

    // TODO: check mx_imgs - cell
    // TODO: check mx_scales - single vector, length same as mx_imgs

    int n_imgs = mxGetNumberOfElements(mx_imgs);
    const float * scales = (const float*)mxGetData(mx_scales);
       
    //mexPrintf("%d images\n",n_imgs);
    Detector * D = createDetectorFromMxArray(rhs[2], rhs[3], rhs[4], rhs[5]);
    if (!D)
    {
        mexErrMsgTxt("Invalid detector");
    }
    //mexPrintf("Detector: %dx%d with %d stages\n", D.w, D.h, D.T);
    
    unsigned int nThreads = 0;
    if (nrhs == 7) nThreads = mxGetScalar(rhs[6]);
    
    vector<BB> bbs;
    vector<float> hs;
    int nf = 0; int ns = 0;
    double ce = 0.0;

    if (nThreads <= 1)
    {
        // No threads version
        list<Image> ims;
        for (int k = 0; k < n_imgs; ++k)
        {
            const mxArray * img_k = mxGetCell(mx_imgs, k);
            // TODO: check img_k - single matrix
            ims.push_back(Image((float*)mxGetData(img_k), mxGetNumberOfDimensions(img_k), mxGetDimensions(img_k)));
            Image & im = ims.back();
            if (!im.data) break;
            const int max_d0 = im.dims[0] - D->wdim0 - 1;
            const int max_d1 = im.dims[1] - D->wdim1 - 1;
            if (max_d0 <= 0 || max_d1 <= 0) break;
            //mexPrintf("Image: %dx%d px, sacle=%.2f, k=%d\n", im.dim0, im.dim1, scales[k], k);
            ScalarResult r = scalarScanImage(std::cref(im), 0, 0, max_d0, max_d1, 1, std::cref(*D), scales[k], true);
            bbs.insert(bbs.end(), r.bbs.begin(), r.bbs.end());
            hs.insert(hs.end(), r.hs.begin(), r.hs.end());
            ns = r.ns; nf = r.nf; ce = r.cacheEffeciency;
        }
    }
    else
    {
        // Threaded version
        nThreads = std::max(1U, std::min(std::thread::hardware_concurrency(), nThreads));
        ThreadPool pool(nThreads);
        const int tile_sz = 128;
        list<Image> ims;
        list< future<ScalarResult> > results;
        for (int k = 0; k < n_imgs; ++k)
        {
            const mxArray * img_k = mxGetCell(mx_imgs, k);
            // TODO: check img_k - single matrix
            ims.push_back(Image((float*)mxGetData(img_k), mxGetNumberOfDimensions(img_k), mxGetDimensions(img_k)));
            const Image & im = ims.back();
            if (!im.data) break;
            const int max_d0 = im.dims[0] - D->wdim0 - 1;
            const int max_d1 = im.dims[1] - D->wdim1 - 1;
            if (max_d0 <= 0 || max_d1 <= 0) break;
            //mexPrintf("Image: %dx%d px, sacle=%.2f, k=%d\n", im.dims[0], im.dims[1], scales[k], k);
            for (int d0 = 0; d0 < max_d0; d0+=tile_sz)
                for (int d1 = 0; d1 < max_d1; d1+=tile_sz)
                {
                    //mexPrintf("x=%i:%i; y=%i:%i\n",x, min(x+TILE_SZ-1, max_x), y, min(y+TILE_SZ-1, max_y));
                    results.emplace_back(
                        pool.enqueue(
                            scalarScanImage, std::cref(im), d0, d1, min(d0+tile_sz, max_d0), min(d1+tile_sz, max_d1), 1, std::cref(*D), scales[k], true // TODO: parameterize stride
                        )
                    );
                }
        }
        for (auto && result: results)
        {
            ScalarResult r = result.get();
            bbs.insert(bbs.end(), r.bbs.begin(), r.bbs.end());
            hs.insert(hs.end(), r.hs.begin(), r.hs.end());
            //mexPrintf("ns=%d, nf=%d, passed: %d\n", r.ns, r.nf, r.bbs.size());
            ns += r.ns; nf += r.nf; ce += r.cacheEffeciency;
        }
        ce /= results.size();
    }
    
    destroyDetector(&D);
    
    // Convert bb to Lx5 matlab matrix with bounding boxes
    lhs[0] = mxCreateNumericMatrix(6, bbs.size(), mxDOUBLE_CLASS, mxREAL);
    lhs[1] = mxCreateDoubleScalar(double(ns));
    lhs[2] = mxCreateDoubleScalar(double(nf));
    lhs[3] = mxCreateDoubleScalar(ce);
    
    double * bbs_out = (double*)mxGetData(lhs[0]);
    for (int i = 0; i < bbs.size(); ++i)
    {
        BB & b = bbs[i];
        float h = hs[i];
        bbs_out[6*i+0] = b.y;
        bbs_out[6*i+1] = b.x;
        bbs_out[6*i+2] = b.height;
        bbs_out[6*i+3] = b.width;
        bbs_out[6*i+4] = h;
        bbs_out[6*i+5] = b.scale;
        //mexPrintf("%.2f, ", h);
    }
    
    return;
}
